/*
__________________________________________________________________________________________

  Name			:	Image.cpp
  Version		:	1.01
  Description	:	Image storage and management. Uses intel JPEG library to load jpeg 
					files. Supports embedded files (except JPEG currently). 

					Several operations included for combining and multiplying images. 
					
					This module is required by Texture v2.0

  Last Update	:	20/03/02

__________________________________________________________________________________________

/*
(c)2002 Ben Woodhouse

ben@elf-stone.com
www.elf-stone.com

This code may not be used in a commercial product without my permission.
If you redistribute it, this message must remain intact. 
*/

//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "image.h" 
#include "vector.h"
#include "jpeglib.h" 

#include <math.h>


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CImage::CImage()
{ 
	memSize=0;
	data=NULL;
}


CImage::CImage(const CImage &src)
{
	data=NULL;
	copyFrom(src); 

}//copy constructor (required by STL)

CImage::~CImage()
{
	if (data!=NULL) delete [] data; 
}

const CImage& CImage::operator= (const CImage& src)
{
	copyFrom(src);
	return *this;
}//(required by STL)


void CImage::copyFrom(const CImage & src)
{
	if (data!=NULL) delete [] data;
	allocate(src.width,src.height,src.numChannels);
	format=src.format;

	//copy the data
	for (int a=0;a<memSize;++a)
	{
		data[a]=src.data[a];
	}
}

bool CImage::convertToNormalMap(bool wrap, float scale)
{
	float oneOver255=1.0/255.0;
	float oneOver127=1.0/127.5;

	if (numChannels!=1) return false; 

	//copy the heightfield into a grid of floats which has the image's dimensions
	//+2 and takes into account wrapping

	float (*nmap)[3];
	float * fmap;

	int widthf=width+2;
	int heightf=height+2;
	int numfTexels=widthf*heightf;
	

	fmap = new float[numfTexels];
	nmap = new float[numfTexels][3];

	int fpos=widthf; //because y starts at 1
	int dpos=0;

	int x,y;

	for (y=1;y<heightf-1;++y)
	{
		++fpos; //because x starts at 1
		for (x=1;x<widthf-1;++x)
		{
			fmap[fpos]=(float)data[dpos]*oneOver255*scale;

			++fpos;
			++dpos;
		}
		++fpos;//skipped x
	}

	//copy source offsets for edge texels
	//right, left, middle and bottom 
	int roffset;
	int loffset;
	int toffset;
	int boffset;

	if (wrap)
	{
		roffset=-widthf+2;
		loffset=widthf-2;
		toffset=-widthf*(heightf-2);
		boffset=widthf*(heightf-2);
	}
	else
	{
		roffset=-1;
		loffset=1;
		toffset=-widthf;
		boffset=widthf;
	}

	//generate the edges depending on the wrap mode
	//ignore corners (they won't be sampled anyway)
	int dsti=1; 
	int srci=dsti+boffset;

	//bottom row
	for (x=1;x<widthf-1;++x)
	{
		fmap[dsti]=fmap[srci];
		++dsti;
		++srci;
	}
	++dsti;

	//middle rows
	for (y=1;y<heightf-1;++y)
	{
		//left side
		fmap[dsti]=fmap[dsti+loffset];

		//right side
		dsti+=widthf-1;
		fmap[dsti]=fmap[dsti+roffset];
		++dsti;
	}

	//top row
	srci=dsti+toffset;
	for (x=1;x<widthf-1;++x)
	{
		fmap[dsti]=fmap[srci];
		++srci;
		++dsti;
	}


	float sTan[4][3]; //top, right, bottom, left
	float sNorm[4][3]; //topright, bottomright, bottomleft, topleft
	//get vectors to 4 surrounding texels
	// +
	//get 4 cross products (one from each pair of adjacent surrounding texels)
	//add them 
	//normalise

	int i=widthf;
	int npos=0;
	for (y=1;y<heightf-1;++y)
	{
		++i; //texel skipped at start
		for (x=1;x<widthf-1;++x)
		{
			//get the tangents to the 4 surrounding texels
			sTan[0][0]=0; //above
			sTan[0][1]=1;
			sTan[0][2]=fmap[i+widthf]-fmap[i]; //above

			sTan[1][0]=1; 
			sTan[1][1]=0;
			sTan[1][2]=fmap[i+1]-fmap[i];		//right

			sTan[2][0]=0; 
			sTan[2][1]=-1;
			sTan[2][2]=fmap[i-widthf]-fmap[i]; //below

			sTan[3][0]=-1; 
			sTan[3][1]=0;
			sTan[3][2]=fmap[i-1]-fmap[i];		//left

			//get 4 normals by finding the cp of the adjacent tangent pairs
			
			crossProduct(sNorm[0],sTan[1],sTan[0]);
			crossProduct(sNorm[1],sTan[2],sTan[1]);
			crossProduct(sNorm[2],sTan[3],sTan[2]);
			crossProduct(sNorm[3],sTan[0],sTan[3]);

			nmap[npos][0]=0;
			nmap[npos][1]=0;
			nmap[npos][2]=0;

			//add the 4 vectors and normalise
			for (int a=0;a<4;++a)
			{
				for (int b=0;b<3;++b)
				{
					nmap[npos][b]+=sNorm[a][b];					
				}
			}
			normalize(nmap[npos]);
			
			++npos;
			++i;
		}
		++i; //texel skipped at end
	}



	//delete the old data 
	delete [] data;

	//allocate some new data
	allocate(width,height,3);

	//put the normal map into the data (mapped from -1:1 to 0:255)
	i=0;
	npos=0;
	for (y=0;y<height;++y)
	{
		for (x=0;x<width;++x)
		{
			for (int a=0;a<3;++a)
			{
				data[i]=(nmap[npos][a]+1)*127;
				//data[i]=random(255); 
				++i;
			}
			++npos;
		}
	}

	setFormat(GL_RGB);

	delete [] fmap;
	delete [] nmap;

	return true;
}

bool CImage::NM_normalize()
{
	if (numChannels!=3) return false; 

	int numPixels=width*height;

	float sumSquared;
	float length;

	int datpos=0;
	for (int a=0;a<numPixels;++a)
	{
		sumSquared=0;

		datpos=a*3;

		//find A^2 + B^2 + C^2...
		for (int b=0;b<numChannels;++b)
		{
			sumSquared+=(float)data[datpos+b]*(float)data[datpos+b];
		}

		//get sqrt (length) and divide by it
		length=sqrt(sumSquared)/255;

		if (length<0.001)
		{
			for (b=0;b<numChannels;++b)
			{
				data[datpos+b]=255;
			}
		}
		else
		{
			for (b=0;b<numChannels;++b)
			{
				data[datpos+b]/=length;
			}
		}


	}
	return true;

}

bool CImage::duplicateChannel(int numChannels_in)
{
	if (numChannels!=1) 
	{
		messageBox("Can't duplicate a multi-channel image");
		return false;
	}
	if (numChannels_in!=3 && numChannels_in!=4) 
	{
		messageBox("Number of destination channels must be 3 or 4");
		return false;
	}

	int newMemSize=numChannels_in*width*height;
	
	//create an appropriately sized array in which to store the new data
	BYTE * newData=new BYTE[newMemSize];

	int numPixels=width*height;

	int count=0;
	for (int a=0;a<numPixels;++a)
	{
		for (int b=0;b<numChannels_in;++b)
		{
			newData[count]=data[a];
			++count;
		}
	}

	numChannels=numChannels_in;
	memSize=newMemSize;

	delete [] data;
	data=newData;

	setLikelyFormat();

	return true;
}
 
BYTE * CImage::getData()
{
	return data;
}

bool CImage::setFormat(GLenum format_in)
{
	format=format_in;

	return true;
}

void CImage::setLikelyFormat()
{
	switch (numChannels)
	{
		case 1:
			format=GL_ALPHA;
			break;

		case 3:
			format=GL_RGB;
			break;

		case 4:
			format=GL_RGBA;
			break;

	}
}

GLenum CImage::getFormat()
{
	return format;
}

int CImage::getNumChannels()
{
	return numChannels;
}

int CImage::getWidth()
{
	return width;
}

int CImage::getHeight()
{
	return height;
}


bool CImage::isPowerOf2(int num)
{
	int testNum=1;

	//test against some testnums up to 2^16
	for (int a=0;a<16;++a)
	{
		testNum=testNum<<1;
		if (num==testNum) return true;
	}
	return false;
}

bool CImage::loadRAW(const char * fn)
{
	FILE * filePtr;
	if ((filePtr = fopen(fn,"rb"))==NULL) 
	{
		messageBox("Image::LoadRAW() Error \n%s not found\n",fn);
		return false;
	}


	bool rv=loadRAWEmbedded(filePtr);

	fclose(filePtr);
	return rv;
}

bool CImage::loadBMP(const char * fn)
{
	FILE * filePtr;
	if ((filePtr = fopen(fn,"rb"))==NULL) 
	{
		messageBox("Image::LoadBMP() Error \n%s not found\n",fn);
		return false;
	}
	bool rv=loadBMPEmbedded(filePtr);

	fclose(filePtr);
	return rv;
}





















bool CImage::loadJPG(const char * fn)
{
	FILE * file;
	if ((file=fopen(fn,"rb"))==NULL) 
	{
		messageBox("Jpg file not found : %s",fn);
		return false;
	}
	bool rv=loadJPGEmbedded(file);

	messageBox("data[1] %d",data[1]);
	
	fclose(file);
	return true;
}

bool CImage::loadJPGEmbedded(FILE * filePtr)
{

/*BUG:
	If you wish to read a series of JPEG images from
a single source file, you should repeat the jpeg_read_header() to
jpeg_finish_decompress() sequence without reinitializing either the JPEG
object or the data source module; this prevents buffered input data from
being discarded.
*/
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, filePtr);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);

	/*
	width=cinfo.output_width;
	height=cinfo.output_height;
	numChannels=3;
	*/
	width=cinfo.image_width;
	height=cinfo.image_width;
	numChannels=cinfo.num_components;

	//allocate(cinfo.image_width,cinfo.image_height,cinfo.num_components);
	allocate();
	int rowSpan=width*cinfo.num_components;
	//messageBox("%d",rowSpan);

	BYTE ** scanPtr=new BYTE *[height];
	for (int a=0;a<height;++a)
	{
		scanPtr[a]=&data[a*rowSpan];
	}

	int rowCount=0;
	while (cinfo.output_scanline<cinfo.output_height) 
	{
		// Read in the current row of pixels and increase the rowsRead count
		rowCount+=jpeg_read_scanlines(&cinfo,&scanPtr[rowCount],cinfo.output_height-rowCount);
	}
	
	jpeg_finish_decompress(&cinfo);
	delete [] scanPtr;

	jpeg_destroy_decompress(&cinfo);

	setLikelyFormat();
	return true;
}
/*
bool CImage::loadJPG(const char * fn)
{
	JPEG_CORE_PROPERTIES jprop;

	BYTE * texdat=NULL;
	
//	BYTE temp;
	int tpos,bpos;

	int x,y;

	int byteSize;
	int lineSize;
	int padBytes;
	
	
	

	if(IJL_OK != ijlInit(&jprop))
	{
		glPrintf("JPG load error: failed to init\n");
		goto error;
	}

	jprop.JPGFile = fn;

	if(ijlRead(&jprop,IJL_JFILE_READPARAMS)!=IJL_OK)
	{
		goto error;
	}

	jprop.DIBWidth  =  jprop.JPGWidth;
	jprop.DIBHeight = jprop.JPGHeight;

	switch(jprop.JPGChannels)
	{
		case 1:
			jprop.DIBChannels = 3;
			jprop.DIBColor    = IJL_RGB;
		break;

		case 3:
			jprop.DIBChannels = 3;
			jprop.DIBColor    = IJL_RGB;
		break;

		case 4:
			jprop.DIBChannels = jprop.JPGChannels;
			jprop.DIBColor    = IJL_RGBA_FPX;
		break;
	}

	width=jprop.DIBWidth;
	height=jprop.DIBHeight;
	numChannels=jprop.JPGChannels;

	padBytes = IJL_DIB_PAD_BYTES(jprop.DIBWidth,jprop.DIBChannels);
	lineSize = jprop.DIBWidth * jprop.DIBChannels + padBytes;
	byteSize = lineSize * abs(jprop.DIBHeight);// + sizeof(BITMAPINFO);
   //m_imageData = new BYTE [wholeSize];

	allocate();

	texdat=new BYTE[byteSize]; 

	jprop.DIBBytes    = texdat;
	jprop.DIBPadBytes = padBytes;


	if(ijlRead(&jprop,IJL_JFILE_READWHOLEIMAGE)!=IJL_OK)
	{
		glPrintf("JPG load error: failed to read stuff\n");
		goto error;
	}

	ijlFree(&jprop);

	//flip texdat over by swapping rows
	for (y=0;y<height;++y)
	{
		//get mem positions for start of current row and its opposite in y 
		tpos=y*width*numChannels;
		bpos=((height-y)-1)*width*numChannels;

		for (x=0;x<width;++x)
		{
 
			for (int a=0;a<numChannels;++a)
			{
				data[tpos]=texdat[bpos];
				tpos++;
				bpos++;
			}
		}

	}

	setLikelyFormat();

	delete [] texdat;
	return true;


error:
	messageBox("JPG load error");
	ijlFree(&jprop);

	if (texdat!=NULL) delete [] texdat;

	return false;

}
*/
//raw files must have equal width and height and either 1 or 3 channels
bool CImage::loadRAWEmbedded(FILE * filePtr)
{
	long embeddedFileStart=ftell(filePtr);

	//get the file size
	fseek(filePtr,0,SEEK_END);
	int fileSize=ftell(filePtr);

	fseek(filePtr,0,SEEK_SET);

	//calculate the number of channels (c), and the width/height (s) from the file size (f)

	int numPixels;

	//if the filesize is divisible by 3 then assume three channels
	if (fileSize%3==0) 
	{
		numPixels=fileSize/3;
		numChannels=3;
	}
	else 
	{
		numPixels=fileSize;
		numChannels=1;
	}

	int size=sqrt(numPixels);

	//check the calculated size is valid
	if (!isPowerOf2(size))
	{
		messageBox("Error: RAW image size is incorrect %d",numPixels);
		return false;
	}

	width=size;
	height=size;

	allocate();

	int tx,ty,datpos=0,pixelpos;//,intex;

	for (ty=size-1;ty>-1;ty--)
	{
		for (tx=0;tx<size;tx++)
		{			
//			modified to standardise. 0Y= BOTTOM
			pixelpos=tx+ty*size;
			datpos=pixelpos*numChannels;

			for (int a=0;a<numChannels;++a)
			{
				data[datpos]=fgetc(filePtr);
				++datpos;
			}
		}
	}
	setLikelyFormat();
	return true;
}

bool CImage::loadBMPEmbedded(FILE * filePtr)
{

	long embeddedFileStart=ftell(filePtr);

	BMPHeader header;
	header.readFromFile(filePtr);

	if (header.id[0]!='B' || header.id[1]!='M') 
	{
		messageBox("CImage::loadBMPEmbedded error: \nFile is not BMP format");
		return false;
	}

	numChannels=header.bpp>>3;
	height=header.height;
	width=header.width;
	int alignBytes=header.getAlignBytes();
	allocate();


	/*
	//check the file is BMP format
	if (!(getc(filePtr)=='B' && getc(filePtr)=='M'))
	{
		messageBox("CImage::loadBMPEmbedded error: \nFile is not BMP format");
		return false;
	}

	 //get the bit depth. If it's 8-bit, it's assumed the data is 
	 //grayscale, not paletted
	fseek(filePtr,28+embeddedFileStart,SEEK_SET);

	int bitDepth=getc(filePtr);

	numChannels=bitDepth>>3;

	//get the offset for the data start
	fseek(filePtr,10+embeddedFileStart,SEEK_SET);
	datStart=_getw(filePtr);

	//get the size in pixels width and height
	(void)_getw(filePtr);
	height=_getw(filePtr);
	width=_getw(filePtr);

	//get the filler bytes
	int alignBytes=4-(width%4);

	if (alignBytes==4) alignBytes=0;

*/

	//set the stream position to the start of the bitmap data
	fseek(filePtr,header.dataOffset+embeddedFileStart,SEEK_SET);

	int tx,ty,datpos,pixelpos=0;
	//special cases
	for (ty=0;ty<height;ty++)
	{
		for (tx=0;tx<width;tx++)
		{
			datpos=pixelpos*numChannels+numChannels-1;
			++pixelpos;			

			for (int a=0;a<numChannels;++a)
			{
				data[datpos]=fgetc(filePtr);
				--datpos;
			}
		}
		for (int a=0;a<alignBytes;++a)
		{
			(void)fgetc(filePtr);
		}
	}
	setLikelyFormat();

	return true;
}

//multiplies the specified channels by a single value
void CImage::multiply(int channelFlags,float amount)
{
	bool channelEnabled[4]={false,false,false,false};

	channelEnabled[0]=(channelFlags & CHANNEL_RED)!=0;
	channelEnabled[1]=(channelFlags & CHANNEL_GREEN)!=0;
	channelEnabled[2]=(channelFlags & CHANNEL_BLUE)!=0;
	channelEnabled[3]=(channelFlags & CHANNEL_ALPHA)!=0;

	int numPixels=width*height;
	int count=0;
	for (int a=0;a<numPixels;++a)
	{
		for (int b=0;b<numChannels;++b)
		{
			if (channelEnabled[b])
				data[count]=data[count]*amount;

			++count;
		}
	}
}

//this can only be done to 3 channel (RGB) images
//the alphaImage must have only one channel
bool CImage::addAlphaChannel(CImage & alphaImage)
{
	if (numChannels!=3) 
	{
		messageBox("Error: Number of channels in original image should be 3. \nCurrently %d",numChannels);
		return false;
	}
	if (alphaImage.getNumChannels()!=1) 
	{
		messageBox("Error: Number of channels in alpha image should be 1. Currently %d",alphaImage.getNumChannels());
		return false;
	}

	int newMemSize=4*width*height;
	int numPixels=width*height;

	BYTE * alphaData=alphaImage.getData();
	
	//create an appropriately sized array in which to store the new data
	BYTE * newData=new BYTE[newMemSize];

	int oi=0; //old index
	int ni=0; //new index
	//for each pixel
	for (int a=0;a<numPixels;++a)
	{
		//add the rgb from the old texture
		for (int b=0;b<3;++b)
		{
			newData[ni]=data[oi];
			++oi;
			++ni;
		}
		//add the alpha component from the other alpha image
		//the alpha data's index is simply the pixel number
		newData[ni]=alphaData[a];
		++ni;
	}

	numChannels=4;
	memSize=newMemSize;

	delete [] data;
	data=newData;

	setLikelyFormat();

	return true;
}

//bool CImage::loadJPGEmbedded(FILE * filePtr);

void CImage::allocate(int width_in, int height_in, int numChannels_in)
{
	if (data!=NULL) delete [] data;
	width=width_in;
	height=height_in;
	numChannels=numChannels_in;
	memSize=width*height*numChannels;
	data=new BYTE[memSize]; 
}

void CImage::allocate()
{
	if (data!=NULL) delete [] data;
	memSize=width*height*numChannels;
	data=new BYTE[memSize]; 
}

//additive blend
void CImage::regionPlusImg(int originx,int originy,CImage &imageb)
{
	//CImage * destImg=new CImage;
	int bright=2;

	int destHeight=imageb.getHeight();
	int destWidth=imageb.getWidth();
	int destChan=imageb.getNumChannels();

	//destImg->allocate(destWidth,destHeight,destChan);

	//destImg->format=imageb.getFormat();

	BYTE * bData=imageb.getData();

//	int dat;
	int mSize=imageb.getMemSize();
//	int amt;

	int x,y;
//	BYTE temp;

//	int ecxi;

	int bigPos=((originx+originy*width)*3)>>2;
	int lilPos=0;

	int bigWidthDif=(width-destWidth)*3;
	int bigWidthDif32=((width-destWidth)>>2)*3;
	int destWidth32=destWidth>>2;

	BYTE * destData=data;//+((originx+originy*width)*3);

	unsigned int * destData32=(unsigned int *)data+bigPos;
	unsigned int * bData32=(unsigned int *)bData;
	int dwTimes3=destWidth32*3;

	int dwt3=destWidth*3;
	int p=0;
/*
	if (bright!=1)
	{
		for (y=0;y<destHeight;++y)
		{
//			for (x=0;x<dwTimes3;++x)
			for (x=0;x<dwt3;++x)
			{
				//*bData32=*bData32&0x88888888;
				//bData32++;
				//*bData<<=1;
		//		bData++;
				bData[p++]>>=1;
			}
		}
		//bData32=(unsigned int *) bData;
	}
*/

/*
	int res;
	//unsigned int * aData32=(unsigned int *)data+bigPos;
	for (y=0;y<destHeight;++y)
	{
		for (x=0;x<dwt3;++x)
		{
			res= *(destData)+ *(bData++);
			if (res>255) *(destData++)=255;
			else *(destData++)=res;
		}
	}
*/
//	int res;

	for (y=0;y<destHeight;++y)
	{ 
		for (x=0;x<dwTimes3;++x)
		{
			*destData32=((*destData32&0xFEFEFEFE)>>1) + (*bData32>>1);

			if (*destData32 & 0x80808080)
			{
				if (*destData32 & 0x80000000)
				{
					*destData32|= 0x7F000000;
				}
				if (*destData32 & 0x00800000)
				{
					*destData32|= 0x007F0000;
				}
				if (*destData32 & 0x00008000)
				{
					*destData32|= 0x00007F00;
				}
				if (*destData32 & 0x00000080)
				{
					*destData32|= 0x0000007F;
				}
				
				*destData32&=0x7F7F7F7F;
			}
			*destData32<<=1;
			destData32++;
			bData32++;
		}
		destData32+=bigWidthDif32;
	}
	
}


	/*
	FAST AVERAGE BLEND
	int res;
	for (y=0;y<destHeight;++y)
	{
		for (x=0;x<destWidth32;++x)
		{
			destData32[lilPos]=(aData32[bigPos++]>>1) + (bData32[lilPos]>>1);
			++lilPos;
			destData32[lilPos]=(aData32[bigPos++]>>1) + (bData32[lilPos]>>1);
			++lilPos;
			destData32[lilPos]=(aData32[bigPos++]>>1) + (bData32[lilPos]>>1);
			++lilPos;

		}
		bigPos+=bigWidthDif32; 
	}
*/

void CImage::copyRegionFrom(CImage & src,int originx, int originy, int rwidth, int rheight)
{
	int datOffset=(originx+originy*width)*3;
	
	BYTE * destDat=&data[datOffset];
	BYTE * srcDat=(src.getData()+datOffset);
	unsigned int * destDat32=(unsigned int *)destDat;
	unsigned int * srcDat32=(unsigned int *)srcDat;
	int rwidth32=rwidth>>2;

	int widthDif32=((width>>2)-(rwidth>>2))*3;

	for (int y=0;y<rheight;++y)
	{
		for (int x=0;x<rwidth32;++x)
		{
			*destDat32 = *srcDat32;
			destDat32++;
			srcDat32++; 
			*destDat32 = *srcDat32;
			destDat32++;
			srcDat32++; 
			*destDat32 = *srcDat32;
			destDat32++;
			srcDat32++; 
		}
		srcDat32+=widthDif32;
		destDat32+=widthDif32;
	}

}

void CImage::copyRegionFromNI(	CImage & src,
								int srcX,
								int srcY,
								int dstX,
								int dstY,
								int rwidth,
								int rheight)
{
	int srcDatOffset=(srcX+srcY*src.width)*3;
	int dstDatOffset=(dstX+dstY*width)*3;
	
	BYTE * destDat=&data[dstDatOffset];
	BYTE * srcDat=(src.getData()+srcDatOffset);

	unsigned int * destDat32=(unsigned int *)destDat;
	unsigned int * srcDat32=(unsigned int *)srcDat;
	int rwidth32=rwidth>>2;

	int dstWidthDif32=((width>>2)-(rwidth>>2))*3;
	int srcWidthDif32=((src.width>>2)-(rwidth>>2))*3;

	for (int y=0;y<rheight;++y)
	{
		for (int x=0;x<rwidth32;++x)
		{
			*destDat32 = *srcDat32;
			destDat32++;
			srcDat32++; 
			*destDat32 = *srcDat32;
			destDat32++;
			srcDat32++; 
			*destDat32 = *srcDat32;
			destDat32++;
			srcDat32++; 
		}
		srcDat32+=srcWidthDif32;
		destDat32+=dstWidthDif32;
	}
}

void CImage::removeLSB()
{
	for (int a=0;a<memSize;++a)
	{
		data[a] = data[a] & 254;
	}
}


bool CImage::resample(int newWidth, int newHeight)
{
	int newChannels=numChannels;
	CImage newImage;
	newImage.allocate(newWidth,newHeight,newChannels);

	BYTE * dat=newImage.getData();

	int p=0;
	for (int y=0;y<newHeight;++y)
	{
		for (int x=0;x<newWidth;++x)
		{
			for (int a=0;a<numChannels;++a)
			{
				dat[p]=getPixelAt_linear( a, (float)x/(float)(newWidth-1.0), (float)y/(float)(newHeight-1.0) );
				++p;
			}
		}
	}

	this->copyFrom(newImage);
	
	return true;
}



bool CImage::resize(int newWidth, int newHeight)
{
	int newChannels=numChannels;
	CImage newImage;
	newImage.allocate(newWidth,newHeight,newChannels);

	BYTE * dat=newImage.getData();

	int p=0;
	for (int y=0;y<newHeight;++y)
	{
		for (int x=0;x<newWidth;++x)
		{
			for (int a=0;a<numChannels;++a)
			{
				dat[p]=getPixelAt_nearest( a, (float)x/(float)(newWidth-1.0), (float)y/(float)(newHeight-1.0) );
				++p;
			}
		}
	}
	this->copyFrom(newImage);

	return true;
}




BYTE CImage::getPixelAt_nearest(int channel, float x, float y)
{
	int xint=x*(float)(width-1);
	int yint=y*(float)(height-1);

	return data[(xint+yint*width)*numChannels+channel];
}

BYTE CImage::getPixelAt_linear(int channel, float x, float y)
{
	//int xint=x*(float)(width-1);
	//int yint=y*(float)(height-1);
	float xpos=x*(float)(width-1);
	float ypos=y*(float)(height-1);

	float oneOver255=1.0/255.0;

	int cornerPos[4][2];
	float cornerHeights[4];
	float cornerDist[4];

	cornerPos[0][0]=(int)xpos;
	cornerPos[0][1]=(int)ypos;
	cornerPos[1][0]=(int)xpos+1;
	cornerPos[1][1]=(int)ypos; 
	cornerPos[2][0]=(int)xpos;
	cornerPos[2][1]=(int)ypos+1;
	cornerPos[3][0]=(int)xpos+1;
	cornerPos[3][1]=(int)ypos+1;

	for (int a=0;a<4;++a)
	{
		int offset=cornerPos[a][1]*width+cornerPos[a][0];

		cornerHeights[a]=(float)data[ offset*numChannels+channel ]*oneOver255;
	}

	float dx,dz;
	for (a=0;a<4;++a)
	{
		dx=xpos-(float)cornerPos[a][0];
		dz=ypos-(float)cornerPos[a][1];

		cornerDist[a]=sqrt(dx*dx + dz*dz);
	}

	float totalHeight=0;
	float cornerWeight[4];
	float totalWeight=0;
	for (a=0;a<4;++a)
	{
		if (cornerDist[a]<=1)
		{
			totalHeight+=cornerHeights[a];
			cornerWeight[a]=1.0-cornerDist[a];
			totalWeight+=cornerWeight[a];
		}
	}

	//get the weighted average
	float weight;
	float heightOut=0;
	for (a=0;a<4;++a)
	{
		if (cornerDist[a]<=1)
		{
			weight=cornerWeight[a]/totalWeight;
			heightOut+=weight*cornerHeights[a];
		}
	}
	//retun heigh

	return (BYTE)(heightOut*255.0);

	//return data[(xint+yint*width)*numChannels+channel];
}

void CImage::adjustBrightness(float brightFactor)
{
	int res;
	for (int a=0;a<memSize;++a)
	{
		res=data[a]*=brightFactor;
		if (res>255) data[a]=255;
		else data[a]=res;
	}
}

void CImage::adjustChannelsBrightness(int channels, float brightFactor)
{
	int res;
	int channel=0;
	for (int a=0;a<memSize;++a)
	{
		if (1<<channel & channels)
		{
			res=data[a]*=brightFactor;
			if (res>255) data[a]=255;
			else data[a]=res;
		}
		++channel;
		if (channel==numChannels) channel=0;
	}
}



//alpha blend
void CImage::regionAlphaBlendImg(int originx,int originy,CImage &imageb)
{
	int bright=2;

	int destHeight=imageb.getHeight();
	int destWidth=imageb.getWidth();
	int destChan=imageb.getNumChannels();

	//int srcWidth=width*4;
	//destImg->allocate(destWidth,destHeight,destChan);

	//destImg->format=imageb.getFormat();


//	int dat;
	int mSize=imageb.getMemSize();

	int x,y;

	int lilPos=((originx+originy*width)*3);

	//int bigWidthDif=(width-destWidth)*3;
	BYTE * destData=data+lilPos;//+lilPos;//+((originx+originy*width)*3);
	BYTE * srcData=imageb.getData();


	//destWidth=1;
	//destHeight=1;
	int widthDif=(width-destWidth)*3;


	BYTE colour[5][3]={ {255,255,255},
						{255,0,0},
						{0,255,0},
						{0,0,255},
						{0,0,0}};
	BYTE alpha;
	float oneOver255=1.0/255;
	int c=0;

//	glPrintf("alphie %d %d %d %d\n",destHeight,destWidth,widthDif,height);
	int destWidth32=destWidth<<4;
	BYTE oneMinusAlpha;

/*
	for (y=0;y<destHeight;++y)
	{
		for (x=0;x<destWidth;++x)
		{
			alpha=oneOver255*(*srcData); //alpha
			++srcData;
			oneMinusAlpha=(1.0-alpha);
			(*destData)=alpha*(*srcData) + oneMinusAlpha*(*destData);
			++destData; ++srcData;
			(*destData)=alpha*(*srcData) + oneMinusAlpha*(*destData);
			++destData; ++srcData;
			(*destData)=alpha*(*srcData) + oneMinusAlpha*(*destData);
			++destData; ++srcData;
		}		destData+=widthDif;
	}	
	*/
	
	int *destData32=(int *)destData;
	int *srcData32=(int *)srcData;

	int destWidthDiv4=destWidth>>2;

	for (y=0;y<destHeight;++y)
	{
		
		for (x=0;x<destWidth;++x)
		{
			alpha=*(srcData); //alpha
			oneMinusAlpha=(255-alpha);
			++srcData;
			(*destData)=(alpha*(*srcData) + oneMinusAlpha*(*destData))>>8;
			++destData; 
			++srcData;
			(*destData)=(alpha*(*srcData) + oneMinusAlpha*(*destData))>>8;
			++destData; 
			++srcData;
			(*destData)=(alpha*(*srcData) + oneMinusAlpha*(*destData))>>8;
			++destData; 
			++srcData;
//			++srcData32;
		}
		destData+=widthDif;
		
	}	
}


bool CImage::RGBAToARGB()
{
	if (numChannels!=4) return false;
	int p=0;
	BYTE temp;

	for (int y=0;y<height;++y)
	{
		for (int x=0;x<width;++x)
		{
			//move the channels right
			temp=data[p+3]; //c3 to temp
			data[p+3]=data[p+2]; //c2 to c3
			data[p+2]=data[p+1]; //c1 to c2
			data[p+1]=data[p]; //c0 to c1
			data[p]=temp; //temp (c3) to c0

			p+=4;
		}
	}
	return true;
}


