// ParticleManager.cpp: implementation of the ParticleManager class.
//
//////////////////////////////////////////////////////////////////////
/*
Description:
ParticleManager is a singleton class/object for managing particles in the 
mirage engine. It contains:
	A fixed-size array of particles
	A vector of pointers to particle systems (of objects with base class particleSystem). 

ParticleSystems
	ParticleSystems have to be allocated independently using new, then registered with particleManager. 
	ParticleManager then calls the system's move() and draw() methods each frame.
	ParticleManager is responsible for the deletion of the system. 
		This occurs when:
			all particles in it die
			particleManager is destroyed
	Each ParticleSystem contains vectors of pointers to particles stored within the particleManager

Particles
	Each particle stores 2 id's
		sysID - the index within the system's vector of pointers
		manID - the index within the manager's
	ParticleManager is responsible for sorting and drawing particles. 
*/

/*
(c)2002 Ben Woodhouse

ben@elf-stone.com
www.elf-stone.com

This code may not be used in a commercial product without my permission.
If you redistribute it, this message must remain intact. 
*/


#include "stdafx.h"
#include "ParticleManager.h"
#include "ParticleExplosion.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ParticleManager gParticles;


ParticleManager::ParticleManager()
{
	particleCount=0;
	numSystems=0;
	numParticles=0;
	int lastBlendMode=-1;
	radixSort=new RadixSort(MAX_PARTICLES);
}

ParticleManager::~ParticleManager()
{
	//delete any still active particle systems
	
	//numSystems=particleSystems.size();
	for (int a=0;a<numSystems;++a)
	{
		delete particleSystems[a];
	}
//	numSystems=particleSystems.size();

	delete radixSort;
}

void ParticleManager::reset()
{
	//delete any still active particle systems
	for (int a=0;a<numSystems;++a)
	{
		delete particleSystems[a];
	}
	numSystems=0;
	numParticles=0;
}

void ParticleManager::move()
{
//	glCls();
//	glPrintf("particleCount %d\n",particleCount);
	for (int a=0;a<numSystems;++a)
	{
		particleSystems[a]->move(); 
		particleSystems[a]->cullDeadParticles();
	}

	//cull the remaining systems
	//std::vector<ParticleSystem *>::iterator end;

	
	for (a=0;a<numSystems;++a)
	{
		if (particleSystems[a]->isDead())
		{
			delete particleSystems[a];
			if (particleSystems[a]!=particleSystems[numSystems-1])
			{
				//replace the deleted system pointer with the end one
				particleSystems[a]=particleSystems[numSystems-1];
				--a;
			}

//			particleSystems.pop_back();
			--numSystems;
		}
	}
	/*
	std::vector<ParticleSystem *>::iterator it;
	it=particleSystems.begin();

	int numSys=particleSystems.size();
	bool recheck;
	ParticleSystem * psp;
	while (it!=particleSystems.end())
	{
		psp=*it;
		recheck=false;
		if (psp->isDead())
		{
			ParticleSystem * back=particleSystems.back();
			if (psp!=NULL) 
			{
				delete psp;
				psp=back; 
				back=NULL;
			}
			if (back!=psp)
			{
				particleSystems.pop_back();
				recheck=true;
			}

			--numSystems;
		}
		if (!recheck) ++it;
//		++it;
	}
	*/
//	numSystems=particleSystems.size();
}

void ParticleManager::draw(float camPos[3], float camNorm[3])
{
	sortParticles(camPos,camNorm);

	glPushAttrib(GL_ENABLE_BIT); 

		glEnable(GL_BLEND);
		glDisable(GL_LIGHTING);
		glDepthMask(GL_FALSE);

		standardTex->bind();

/*
		for (int a=0;a<numSystems;++a)
		{
			particleSystems.at(a)->draw();
		}
		*/
		glBlendFunc(GL_ONE, GL_ONE);
		lastBlendMode=-1;
		Particle * part;
		for (int a=numParticles-1;a>=0;--a)
		{
			part=particlep[partOrder[a]];

			if (part->blendMode!=lastBlendMode)
			{
				lastBlendMode=part->blendMode;
				if (lastBlendMode==PBLEND_DARK)
				{
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				}
				else
				{
					glBlendFunc(GL_ONE, GL_ONE);
				}
			}
			part->draw();
		}
		
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthMask(GL_TRUE);
	glPopAttrib();
}

void ParticleManager::sortParticles(float camPos[3],float camNorm[3])
{
	//get the particles
	ParticleSystem * sys;
	numParticles=0;
	bool needsSorting=false;
	int a,b;
	for (a=0;a<numSystems;++a)
	{
		sys=particleSystems[a];
		for (b=0;b<sys->getNumParts();++b)
		{
			particlep[numParticles]=sys->getParticlep(b);
			partDist[numParticles]=particlep[numParticles]->getDistFromCam(camPos,camNorm);
			partOrder[numParticles]=numParticles;
			if (particlep[numParticles]->blendMode==PBLEND_DARK) 
			{
				needsSorting=true;
			}
	
			++numParticles;
			if (numParticles==MAX_PARTICLES)
			{
				goto breakOut; //it's not a goto, it's a double break :P
			}
		}
	}
breakOut:

	//only sorts if there are dark blended (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA) particles
	if (needsSorting)
	{
		radixSort->sort(partDist,numParticles,partOrder);
	}
}

void ParticleManager::registerSystem(ParticleSystem * sys)
{
	//ParticleSystem * newSystem;
	//newList.tex=tex;
	//particleSystems.push_back(sys);
	particleSystems[numSystems]=sys;
	++numSystems;
}

/*
Particle * ParticleManager::createParticle()
{
	if (numParticles>=MAX_PARTICLES)
	{
		return NULL;
	}
	Particle * p=&particles[numParticles++];
	return p;
}

void ParticleManager::destroyParticle(Particle * part)
{
	//swap the last particle into the first particle's position in the array
	--numParticles;
	Particle & lastPart=particles[numParticles];
	*part=lastPart;

	//update the deleted particle's system's pointer

		

	//update the last particle's system's pointer
	lastPart.system->particlep[lastPart.sysID]=part;
}*/
